"""
Test the DSLS parser.
"""

from pytest import mark

from lm_agent.models import LicenseUsesItem, ParsedFeatureItem
from lm_agent.parsing.dsls import parse, parse_feature_dict, parse_usage_dict


@mark.parametrize(
    "feature_dict,parsed_feature",
    [
        (
            {
                "Editor": "Dassault Systemes",
                "EditorId": "5E756A80",
                "Feature": "PAC",
                "Model": "Token",
                "Commercial Type": "STD",
                "Max Release Number": "8",
                "Max Release Date": "2025-01-01 00:59:00",
                "Pricing Structure": "YLC",
                "Max Casual Duration": "0",
                "Expiration Date": "2025-01-01 00:59:00",
                "Customer ID": "11723",
                "Count": "1",
                "Inuse": "0",
                "Tokens": "",
                "Casual Usage (mn)": None,
                "Host": None,
                "User": None,
                "Internal ID": None,
                "Active Process": None,
                "Client Code Version": None,
                "Session ID": None,
                "Granted Since": None,
                "Last Used At": None,
                "Granted At": None,
                "Queue Position": None,
            },
            ParsedFeatureItem(feature="pac", total=1, used=0, uses=[]),
        ),
        (
            {
                "Editor": "Dassault Systemes",
                "EditorId": "5E756A80",
                "Feature": "SRU",
                "Model": "Token",
                "Commercial Type": "STD",
                "Max Release Number": "423",
                "Max Release Date": "2025-01-01 00:59:00",
                "Pricing Structure": "YLC",
                "Max Casual Duration": "0",
                "Expiration Date": "2025-01-01 00:59:00",
                "Customer ID": "11723",
                "Count": "2374",
                "Inuse": "1559",
                "Tokens": "493",
                "Casual Usage (mn)": None,
                "Host": "nid00123 (0123.0)/127.0.0.1",
                "User": "user_1",
                "Internal ID": "SRU",
                "Active Process": "/powerflow/pf_sim_comm ( 48728)",
                "Client Code Version": "6.424",
                "Session ID": "2DBE16",
                "Granted Since": "2024-09-17 17:59:34",
                "Last Used At": "2024-09-18 15:25:50",
                "Granted At": "2024-09-17 17:59:34",
                "Queue Position": None,
            },
            ParsedFeatureItem(feature="sru", total=2374, used=1559, uses=[]),
        ),
        (
            {},
            None,
        ),
    ],
)
def test_parse_feature_dict(feature_dict, parsed_feature):
    """
    Does the parse_feature_dict function extracts the correct data from the csv line?
    The line contains several fields, but we only need:
    - feature
    - count
    - in_use
    """
    assert parse_feature_dict(feature_dict) == parsed_feature


@mark.parametrize(
    "usage_dict,parsed_usage",
    [
        (
            {
                "Editor": "Dassault Systemes",
                "EditorId": "5E756A80",
                "Feature": "PAC",
                "Model": "Token",
                "Commercial Type": "STD",
                "Max Release Number": "8",
                "Max Release Date": "2025-01-01 00:59:00",
                "Pricing Structure": "YLC",
                "Max Casual Duration": "0",
                "Expiration Date": "2025-01-01 00:59:00",
                "Customer ID": "11723",
                "Count": "1",
                "Inuse": "0",
                "Tokens": "",
                "Casual Usage (mn)": None,
                "Host": None,
                "User": None,
                "Internal ID": None,
                "Active Process": None,
                "Client Code Version": None,
                "Session ID": None,
                "Granted Since": None,
                "Last Used At": None,
                "Granted At": None,
                "Queue Position": None,
            },
            None,
        ),
        (
            {
                "Editor": "Dassault Systemes",
                "EditorId": "5E756A80",
                "Feature": "SRU",
                "Model": "Token",
                "Commercial Type": "STD",
                "Max Release Number": "423",
                "Max Release Date": "2025-01-01 00:59:00",
                "Pricing Structure": "YLC",
                "Max Casual Duration": "0",
                "Expiration Date": "2025-01-01 00:59:00",
                "Customer ID": "11723",
                "Count": "2374",
                "Inuse": "1559",
                "Tokens": "493",
                "Casual Usage (mn)": None,
                "Host": "nid00123 (0123.0)/127.0.0.1",
                "User": "user_1",
                "Internal ID": "SRU",
                "Active Process": "/powerflow/pf_sim_comm ( 48728)",
                "Client Code Version": "6.424",
                "Session ID": "2DBE16",
                "Granted Since": "2024-09-17 17:59:34",
                "Last Used At": "2024-09-18 15:25:50",
                "Granted At": "2024-09-17 17:59:34",
                "Queue Position": None,
            },
            LicenseUsesItem(username="user_1", lead_host="nid00123", booked=493),
        ),
        (
            {},
            None,
        ),
    ],
)
def test_parse_usage_dict(usage_dict, parsed_usage):
    """
    Does the parse_usage_dict function extracts the correct data from the csv line?
    The line contains:
    - user name
    - lead host
    - booked
    """
    assert parse_usage_dict(usage_dict) == parsed_usage


def test_parse__correct_output(dsls_output):
    """
    Does the parser return the correct data for this output?
    - dsls_output: expected output from the license server,
    which contain licenses and usage information.
    """
    assert parse(dsls_output) == {
        "pac": ParsedFeatureItem(feature="pac", total=1, used=0, uses=[]),
        "paj": ParsedFeatureItem(feature="paj", total=6, used=0, uses=[]),
        "pca": ParsedFeatureItem(feature="pca", total=4, used=0, uses=[]),
        "pco": ParsedFeatureItem(feature="pco", total=2, used=0, uses=[]),
        "pd5": ParsedFeatureItem(feature="pd5", total=1, used=0, uses=[]),
        "pv6": ParsedFeatureItem(feature="pv6", total=7, used=0, uses=[]),
        "pvg": ParsedFeatureItem(feature="pvg", total=2, used=0, uses=[]),
        "pw7": ParsedFeatureItem(
            feature="pw7",
            total=2000,
            used=2,
            uses=[LicenseUsesItem(username="user_1", lead_host="nid001627", booked=2)],
        ),
        "pw8": ParsedFeatureItem(feature="pw8", total=2000, used=0, uses=[]),
        "sru": ParsedFeatureItem(
            feature="sru",
            total=2374,
            used=1559,
            uses=[
                LicenseUsesItem(username="user_2", lead_host="nid001626", booked=493),
                LicenseUsesItem(username="user_3", lead_host="nid001601", booked=533),
                LicenseUsesItem(username="user_4", lead_host="nid001671", booked=533),
            ],
        ),
    }


def test_parse__bad_output(dsls_output_bad):
    """
    Does the parser return the correct data for this output?
    - dsls_output_bad: unparseable output from the license server,
    which can happen when a connection error occours.
    """
    assert parse(dsls_output_bad) == {}


def test_parse__no_licenses_output(dsls_output_no_licenses):
    """
    Does the parser return the correct data for this output?
    - dsls_output_no_licenses: expected output from the server
    when none of the licenses are in use by users.
    """
    assert parse(dsls_output_no_licenses) == {
        "pac": ParsedFeatureItem(feature="pac", total=1, used=0, uses=[]),
        "paj": ParsedFeatureItem(feature="paj", total=6, used=0, uses=[]),
        "pca": ParsedFeatureItem(feature="pca", total=4, used=0, uses=[]),
        "pco": ParsedFeatureItem(feature="pco", total=2, used=0, uses=[]),
        "pd5": ParsedFeatureItem(feature="pd5", total=1, used=0, uses=[]),
        "pv6": ParsedFeatureItem(feature="pv6", total=7, used=0, uses=[]),
        "pvg": ParsedFeatureItem(feature="pvg", total=2, used=0, uses=[]),
        "pw7": ParsedFeatureItem(feature="pw7", total=2000, used=0, uses=[]),
        "pw8": ParsedFeatureItem(feature="pw8", total=2000, used=0, uses=[]),
        "sru": ParsedFeatureItem(feature="sru", total=2374, used=0, uses=[]),
    }



"""
Test the FlexLM parser.
"""

from pytest import mark

from lm_agent.models import LicenseUsesItem
from lm_agent.parsing.flexlm import parse, parse_feature_line, parse_usage_line


@mark.parametrize(
    "line,result",
    [
        (
            "Users of TESTFEATURE:  (Total of 1000 licenses issued;  Total of 93 licenses in use)",
            {
                "feature": "testfeature",
                "total": 1000,
                "used": 93,
            },
        ),
        (
            "Users of TESTFEATURE:  (Total of 1000 licenses issued;  Total of 1 license in use)",
            {
                "feature": "testfeature",
                "total": 1000,
                "used": 1,
            },
        ),
        (
            "Users of TESTFEATURE:  (Total of 1 license issued;  Total of 1 license in use)",
            {
                "feature": "testfeature",
                "total": 1,
                "used": 1,
            },
        ),
        (
            "not a feature line",
            None,
        ),
        (
            "",
            None,
        ),
    ],
)
def test_parse_feature_line(line, result):
    """
    Does the regex for the feature line match the lines in the output?

    The line contains:
    - feature
    - total
    - used
    """
    assert parse_feature_line(line) == result


@mark.parametrize(
    "line,result",
    [
        (
            "    user1 myserver.example.com /dev/tty (v62.2) (myserver.example.com/24200 12507), start Thu 10/29 8:09, 29 licenses",
            LicenseUsesItem(
                username="user1",
                lead_host="myserver.example.com",
                booked=29,
            ),
        ),
        (
            "    user2 another.server.com /dev/tty feature=feature (v2023.0) (another.server.com/41020 10223), start Mon 3/11 13:16, 100 licenses",
            LicenseUsesItem(
                username="user2",
                lead_host="another.server.com",
                booked=100,
            ),
        ),
        (
            "    user3 ER1234 SESOD5045 MSCONE:ADAMS_View (v2023.0331) (alternative.server.com/29065 2639), start Fri 3/8 13:25, 5 licenses",
            LicenseUsesItem(
                username="user3",
                lead_host="ER1234",
                booked=5,
            ),
        ),
        (
            "    user3 ER1234 SESOD5045 MSCONE:ADAMS_View (v2023.0331) (alternative.server.com/29065 2639), start Fri 3/8 13:25",
            LicenseUsesItem(
                username="user3",
                lead_host="ER1234",
                booked=1,
            ),
        ),
        ("aaaaa", None),
    ],
)
def test_parse_usage_line(line, result):
    """
    Does the regex for the usage line match the line in the output?

    The line contains:
    - user name
    - lead host
    - booked
    """
    assert parse_usage_line(line) == result


@mark.parametrize(
    "fixture,result",
    [
        (
            "flexlm_output",
            {
                "feature": "testfeature",
                "total": 1000,
                "used": 93,
                "uses": [
                    LicenseUsesItem(booked=29, lead_host="myserver.example.com", username="sdmfva"),
                    LicenseUsesItem(booked=27, lead_host="myserver.example.com", username="adfdna"),
                    LicenseUsesItem(booked=37, lead_host="myserver.example.com", username="sdmfva"),
                ],
            },
        ),
        (
            "flexlm_output_2",
            {
                "feature": "test_feature",
                "total": 42800,
                "used": 1600,
                "uses": [
                    LicenseUsesItem(booked=100, lead_host="p-c94.com", username="usbn12"),
                    LicenseUsesItem(booked=1400, lead_host="p-c94.com", username="usbn12"),
                    LicenseUsesItem(booked=100, lead_host="p-c94.com", username="usbn12"),
                ],
            },
        ),
        (
            "flexlm_output_3",
            {
                "feature": "ccmppower",
                "total": 40,
                "used": 3,
                "uses": [
                    LicenseUsesItem(booked=1, lead_host="dcv033.com", username="1nou7p"),
                    LicenseUsesItem(booked=1, lead_host="n-c41.com", username="1nou7p"),
                    LicenseUsesItem(booked=1, lead_host="nid001234", username="1nou7p"),
                ],
            },
        ),
        (
            "flexlm_output_4",
            {
                "feature": "mscone",
                "total": 750,
                "used": 18,
                "uses": [
                    LicenseUsesItem(booked=5, lead_host="ER0037", username="abcdkk"),
                    LicenseUsesItem(booked=1, lead_host="ER0037", username="abcdkk"),
                    LicenseUsesItem(booked=5, lead_host="ER0037", username="abcdkk"),
                    LicenseUsesItem(booked=1, lead_host="ER0037", username="abcdkk"),
                    LicenseUsesItem(booked=5, lead_host="ER0037", username="abcdkk"),
                    LicenseUsesItem(booked=1, lead_host="ER0037", username="abcdkk"),
                ],
            },
        ),
        (
            "flexlm_output_5",
            {
                "feature": "adamstk_gg_solver",
                "total": 6,
                "used": 1,
                "uses": [
                    LicenseUsesItem(booked=1, lead_host="dcv055.com", username="fjan1a"),
                ],
            },
        ),
    ],
)
def test_parse__correct_output(request, fixture, result):
    """
    Does the parser return the correct data for this output?

    - flexlm_output: expected output from the license server,
    which contain licenses and usage information.
    """
    output = request.getfixturevalue(fixture)
    assert parse(output) == result


def test_parse__bad_output(flexlm_output_bad):
    """
    Does the parser return the correct data for this output?

    - flexlm_output_bad: unparseable output from the license server,
    which can happen when a connection error occours.
    """
    assert parse(flexlm_output_bad) == {}


def test_parse__not_licenses_output(flexlm_output_no_licenses):
    """
    Does the parser return the correct data for this output?

    - flexlm_output_no_licenses: expected output from the license server,
    when none of the licenses are in use by users.
    """
    assert parse(flexlm_output_no_licenses) == {
        "feature": "testfeature",
        "total": 1000,
        "used": 0,
        "uses": [],
    }



"""
Test the RLM license server interface.
"""

from unittest import mock

from pytest import fixture, mark, raises

from lm_agent.config import settings
from lm_agent.exceptions import CommandFailedToExecute, LicenseManagerBadServerOutput
from lm_agent.models import LicenseReportItem
from lm_agent.server_interfaces.rlm import RLMLicenseServer


@fixture
def rlm_server(one_configuration_row_rlm) -> RLMLicenseServer:
    return RLMLicenseServer(one_configuration_row_rlm.license_servers)


def test_get_rlm_commands_list(rlm_server: RLMLicenseServer):
    """
    Do the commands for invoking the license server have the correct data?
    """
    commands_list = rlm_server.get_commands_list()
    assert commands_list == [
        [
            f"{settings.RLMUTIL_PATH}",
            "rlmstat",
            "-c",
            "2345@127.0.0.1",
            "-a",
            "-p",
        ]
    ]


@mark.asyncio
@mock.patch("lm_agent.server_interfaces.rlm.run_command")
async def test_rlm_get_output_from_server(
    run_command_mock: mock.MagicMock,
    rlm_server: RLMLicenseServer,
    rlm_output: str,
):
    """
    Do the license server interface return the output from the license server?
    """
    run_command_mock.return_value = rlm_output

    output = await rlm_server.get_output_from_server()
    assert output == rlm_output


@mark.asyncio
@mock.patch("lm_agent.server_interfaces.rlm.RLMLicenseServer.get_output_from_server")
async def test_rlm_get_report_item(
    get_output_from_server_mock: mock.MagicMock, rlm_server: RLMLicenseServer, rlm_output: str
):
    """
    Do the RLM server interface generate a report item for the product?
    """
    get_output_from_server_mock.return_value = rlm_output

    assert await rlm_server.get_report_item(1, "converge.converge_super") == LicenseReportItem(
        feature_id=1,
        product_feature="converge.converge_super",
        used=93,
        total=1000,
        uses=[
            {"username": "asdj13", "lead_host": "myserver.example.com", "booked": 29},
            {"username": "cddcp2", "lead_host": "myserver.example.com", "booked": 27},
            {"username": "asdj13", "lead_host": "myserver.example.com", "booked": 37},
        ],
    )


@mark.asyncio
@mock.patch("lm_agent.server_interfaces.rlm.RLMLicenseServer.get_output_from_server")
async def test_rlm_get_report_item_with_bad_output(
    get_output_from_server_mock: mock.MagicMock, rlm_server: RLMLicenseServer, lsdyna_output_bad: str
):
    """
    Do the RLM server interface raise an exception when the server returns an unparseable output?
    """
    get_output_from_server_mock.return_value = lsdyna_output_bad

    with raises(LicenseManagerBadServerOutput):
        await rlm_server.get_report_item(1, "converge.converge_super")


@mark.asyncio
@mock.patch("lm_agent.server_interfaces.rlm.RLMLicenseServer.get_output_from_server")
async def test_rlm_get_report_item_with_no_used_licenses(
    get_output_from_server_mock: mock.MagicMock, rlm_server: RLMLicenseServer, rlm_output_no_licenses: str
):
    """
    Do the RLM server interface generate a report item when no licenses are in use?
    """
    get_output_from_server_mock.return_value = rlm_output_no_licenses

    assert await rlm_server.get_report_item(1, "converge.converge_super") == LicenseReportItem(
        feature_id=1,
        product_feature="converge.converge_super",
        used=0,
        total=1000,
        used_licenses=[],
    )


@mark.asyncio
@mock.patch("lm_agent.server_interfaces.rlm.run_command")
@mock.patch("lm_agent.server_interfaces.rlm.RLMLicenseServer.get_commands_list")
async def test_rlm_get_report_item_continues_on_exception(
    get_commands_list_mock: mock.MagicMock,
    run_command_mock: mock.MagicMock,
    rlm_server: RLMLicenseServer,
    rlm_output: str,
):
    """
    Do the RLM server interface check the next command if the previous one fails?
    """
    get_commands_list_mock.return_value = [
        [
            f"{settings.RLMUTIL_PATH}",
            "rlmstat",
            "-c",
            "2345@127.0.0.1",
            "-a",
            "-p",
        ],
        [
            f"{settings.RLMUTIL_PATH}",
            "rlmstat",
            "-c",
            "3456@127.0.0.1",
            "-a",
            "-p",
        ],
    ]

    run_command_mock.side_effect = [CommandFailedToExecute("Command failed for first server"), rlm_output]

    assert await rlm_server.get_report_item(1, "converge.converge_super") == LicenseReportItem(
        feature_id=1,
        product_feature="converge.converge_super",
        used=93,
        total=1000,
        uses=[
            {"username": "asdj13", "lead_host": "myserver.example.com", "booked": 29},
            {"username": "cddcp2", "lead_host": "myserver.example.com", "booked": 27},
            {"username": "asdj13", "lead_host": "myserver.example.com", "booked": 37},
        ],
    )



"""
Test the LS-Dyna license server interface.
"""

from unittest import mock

from pytest import fixture, mark, raises

from lm_agent.config import settings
from lm_agent.exceptions import CommandFailedToExecute, LicenseManagerBadServerOutput
from lm_agent.models import LicenseReportItem
from lm_agent.server_interfaces.lsdyna import LSDynaLicenseServer


@fixture
def lsdyna_server(one_configuration_row_lsdyna) -> LSDynaLicenseServer:
    return LSDynaLicenseServer(one_configuration_row_lsdyna.license_servers)


def test_get_lsdyna_commands_list(lsdyna_server: LSDynaLicenseServer):
    """
    Do the commands for invoking the license server have the correct data?
    """
    commands_list = lsdyna_server.get_commands_list()
    assert commands_list == [
        [
            f"{settings.LSDYNA_PATH}",
            "-s",
            "2345@127.0.0.1",
            "-R",
        ]
    ]


@mark.asyncio
@mock.patch("lm_agent.server_interfaces.lsdyna.run_command")
async def test_lsdyna_get_output_from_server(
    run_command_mock: mock.MagicMock,
    lsdyna_server: LSDynaLicenseServer,
    lsdyna_output: str,
):
    """
    Do the license server interface return the output from the license server?
    """
    run_command_mock.return_value = lsdyna_output

    output = await lsdyna_server.get_output_from_server()
    assert output == lsdyna_output


@mark.asyncio
@mock.patch("lm_agent.server_interfaces.lsdyna.LSDynaLicenseServer.get_output_from_server")
async def test_lsdyna_get_report_item(
    get_output_from_server_mock: mock.MagicMock, lsdyna_server: LSDynaLicenseServer, lsdyna_output: str
):
    """
    Do the LS-Dyna server interface generate a report item for the product?
    """
    get_output_from_server_mock.return_value = lsdyna_output

    assert await lsdyna_server.get_report_item(1, "mppdyna.mppdyna") == LicenseReportItem(
        feature_id=1,
        product_feature="mppdyna.mppdyna",
        used=440,
        total=500,
        uses=[
            {"username": "dvds3g", "lead_host": "n-c13.com", "booked": 80},
            {"username": "ssss1d", "lead_host": "n-c52.com", "booked": 80},
            {"username": "ssss1d", "lead_host": "n-c15.com", "booked": 80},
            {"username": "ywap0o", "lead_host": "n-c53.com", "booked": 80},
            {"username": "ywap0o", "lead_host": "n-c51.com", "booked": 80},
            {"username": "ndha1a", "lead_host": "n-c55.com", "booked": 40},
        ],
    )


@mark.asyncio
@mock.patch("lm_agent.server_interfaces.lsdyna.LSDynaLicenseServer.get_output_from_server")
async def test_lsdyna_get_report_item_with_bad_output(
    get_output_from_server_mock: mock.MagicMock, lsdyna_server: LSDynaLicenseServer, lsdyna_output_bad: str
):
    """
    Do the LS-Dyna server interface raise an exception when the server returns an unparseable output?
    """
    get_output_from_server_mock.return_value = lsdyna_output_bad

    with raises(LicenseManagerBadServerOutput):
        await lsdyna_server.get_report_item(1, "mppdyna.mppdyna")


@mark.asyncio
@mock.patch("lm_agent.server_interfaces.lsdyna.LSDynaLicenseServer.get_output_from_server")
async def test_lsdyna_get_report_item_with_no_used_licenses(
    get_output_from_server_mock: mock.MagicMock,
    lsdyna_server: LSDynaLicenseServer,
    lsdyna_output_no_licenses: str,
):
    """
    Do the LS-Dyna server interface generate a report item when no licenses are in use?
    """
    get_output_from_server_mock.return_value = lsdyna_output_no_licenses

    assert await lsdyna_server.get_report_item(1, "mppdyna.mppdyna") == LicenseReportItem(
        feature_id=1,
        product_feature="mppdyna.mppdyna",
        used=0,
        total=500,
        used_licenses=[],
    )


@mark.asyncio
@mock.patch("lm_agent.server_interfaces.lsdyna.run_command")
@mock.patch("lm_agent.server_interfaces.lsdyna.LSDynaLicenseServer.get_commands_list")
async def test_lsdyna_get_report_item_continues_on_exception(
    get_commands_list_mock: mock.MagicMock,
    run_command_mock: mock.MagicMock,
    lsdyna_server: LSDynaLicenseServer,
    lsdyna_output: str,
):
    """
    Do the LS-Dyna server interface check the next command if the previous one fails?
    """
    get_commands_list_mock.return_value = [
        [
            f"{settings.LSDYNA_PATH}",
            "-s",
            "2345@127.0.0.1",
            "-R",
        ],
        [
            f"{settings.LSDYNA_PATH}",
            "-s",
            "3456@127.0.0.1",
            "-R",
        ],
    ]

    run_command_mock.side_effect = [CommandFailedToExecute("Command failed for first server"), lsdyna_output]

    assert await lsdyna_server.get_report_item(1, "mppdyna.mppdyna") == LicenseReportItem(
        feature_id=1,
        product_feature="mppdyna.mppdyna",
        used=440,
        total=500,
        uses=[
            {"username": "dvds3g", "lead_host": "n-c13.com", "booked": 80},
            {"username": "ssss1d", "lead_host": "n-c52.com", "booked": 80},
            {"username": "ssss1d", "lead_host": "n-c15.com", "booked": 80},
            {"username": "ywap0o", "lead_host": "n-c53.com", "booked": 80},
            {"username": "ywap0o", "lead_host": "n-c51.com", "booked": 80},
            {"username": "ndha1a", "lead_host": "n-c55.com", "booked": 40},
        ],
    )



"""
Test Slurm cmd_utils.
"""

from textwrap import dedent
from typing import List
from unittest import mock

from pytest import fixture, mark, raises

from lm_agent.exceptions import ScontrolRetrievalFailure, SqueueParserUnexpectedInputError
from lm_agent.models import LicenseBooking
from lm_agent.workload_managers.slurm.cmd_utils import (
    _match_requested_license,
    get_all_features_cluster_values,
    get_all_product_features_from_cluster,
    get_lead_host,
    get_required_licenses_for_job,
    squeue_parser,
)


@fixture
def job_licenses_good() -> str:
    return "testproduct1.test-feature1@flexlm:11,testproduct2.testfeature2@flexlm:22,testproduct3.testfeature3@flexlm:33"  # noqa: E501


@fixture
def job_licenses_bad() -> str:
    return ""


def test_squeue_with_bad_input():
    """Test that squeue throws the correct exception given incorrect input."""
    with raises(SqueueParserUnexpectedInputError):
        _ = squeue_parser("bad input")


def test_squeue_parser_returns_correct_output_format():
    """Given the squeue formatted output, ensure the `squeue_parsed()` returns the expected values."""
    squeue_parsed_output = [
        {"job_id": 1, "run_time_in_seconds": 300, "state": "RUNNING"},
        {"job_id": 2, "run_time_in_seconds": 184981, "state": "RUNNING"},
        {"job_id": 3, "run_time_in_seconds": 17084, "state": "RUNNING"},
    ]
    squeue_parsed = squeue_parser(
        "\n".join(
            [
                "1|5:00|RUNNING",
                "2|2-3:23:01|RUNNING",
                "3|4:44:44|RUNNING",
            ]
        )
    )
    assert squeue_parsed == squeue_parsed_output


@mark.parametrize(
    "license,output",
    [
        (
            "product.feature@flexlm:123",
            {
                "product_feature": "product.feature",
                "server_type": "flexlm",
                "tokens": 123,
            },
        ),
        (
            "product.feature@rlm",
            {
                "product_feature": "product.feature",
                "server_type": "rlm",
                "tokens": 1,
            },
        ),
        (
            "product.feature-123@flexlm",
            {
                "product_feature": "product.feature-123",
                "server_type": "flexlm",
                "tokens": 1,
            },
        ),
        (
            "product-123.feature@rlm",
            {
                "product_feature": "product-123.feature",
                "server_type": "rlm",
                "tokens": 1,
            },
        ),
    ],
)
def test_match_requested_license(license, output):
    requested_license = license

    return_value = _match_requested_license(requested_license)

    assert return_value == output


@mark.parametrize(
    "requested_license",
    [
        ("productfeature@flexlm:bla"),
        ("productfeature@flexlm:999"),
        ("product.featureflexlm:999"),
        ("productfeatureflexlm999"),
        (""),
        ("product.feature:flexlm@999"),
    ],
)
def test_match_requested_license_wrong_string(requested_license):
    assert _match_requested_license(requested_license) is None


def test_get_required_licenses_for_job_good(job_licenses_good: str):
    """
    Do I return the correct licenses when the license format matches?
    """
    required_licenses = get_required_licenses_for_job(job_licenses_good)
    assert len(required_licenses) == 3
    assert required_licenses == [
        LicenseBooking(
            product_feature="testproduct1.test-feature1",
            quantity=11,
        ),
        LicenseBooking(
            product_feature="testproduct2.testfeature2",
            quantity=22,
        ),
        LicenseBooking(
            product_feature="testproduct3.testfeature3",
            quantity=33,
        ),
    ]


def test_get_required_licenses_for_job_bad(job_licenses_bad: None):
    """
    Do I return the correct licenses when the license format doesn't match?
    """
    required_licenses = get_required_licenses_for_job(job_licenses_bad)
    assert len(required_licenses) == 0
    assert required_licenses == []


@mark.asyncio
@mark.parametrize(
    "show_lic_output,features_from_cluster",
    [
        (
            dedent(
                """
                LicenseName=testproduct1.feature1@flexlm
                    Total=10 Used=0 Free=10 Reserved=0 Remote=yes
                """
            ),
            ["testproduct1.feature1"],
        ),
        (
            dedent(
                """
                LicenseName=product_name.feature_name@flexlm
                    Total=10 Used=0 Free=10 Reserved=0 Remote=yes
                """
            ),
            ["product_name.feature_name"],
        ),
        (
            dedent(
                """
                LicenseName=converge.converge_super@rlm
                    Total=9 Used=0 Free=9 Reserved=0 Remote=yes
                LicenseName=converge.converge_tecplot@rlm
                    Total=45 Used=0 Free=45 Reserved=0 Remote=yes
                """
            ),
            ["converge.converge_super", "converge.converge_tecplot"],
        ),
        (
            dedent(
                """
                LicenseName=powerflow.powerflow-disc@rlm
                    Total=9 Used=0 Free=9 Reserved=0 Remote=yes
                """
            ),
            ["powerflow.powerflow-disc"],
        ),
        ("", []),
    ],
)
@mock.patch("lm_agent.workload_managers.slurm.cmd_utils.scontrol_show_lic")
async def test_get_product_features_from_cluster(
    show_lic_mock: mock.MagicMock, show_lic_output: str, features_from_cluster: List[str]
):
    show_lic_mock.return_value = show_lic_output
    assert features_from_cluster == await get_all_product_features_from_cluster()


@mark.asyncio
@mark.parametrize(
    "show_lic_output,used_features",
    [
        (
            dedent(
                """
                LicenseName=abaqus.abaqus@flexlm
                    Total=1000 Used=90 Free=910 Reserved=0 Remote=yes
                LicenseName=product_name.feature_name@flexlm
                    Total=10 Used=0 Free=10 Reserved=0 Remote=yes
                LicenseName=converge.converge_super@rlm
                    Total=9 Used=1 Free=9 Reserved=0 Remote=yes
                LicenseName=converge.converge_tecplot@rlm
                    Total=45 Used=12 Free=45 Reserved=0 Remote=yes
                """
            ),
            {
                "abaqus.abaqus": {"total": 1000, "used": 90},
                "product_name.feature_name": {"total": 10, "used": 0},
                "converge.converge_super": {"total": 9, "used": 1},
                "converge.converge_tecplot": {"total": 45, "used": 12},
            },
        ),
        (
            "",
            {},
        ),
    ],
)
@mock.patch("lm_agent.workload_managers.slurm.cmd_utils.scontrol_show_lic")
async def test_get_all_features_cluster_values(
    show_lic_mock: mock.MagicMock, show_lic_output: str, used_features: dict
):
    show_lic_mock.return_value = show_lic_output
    assert used_features == await get_all_features_cluster_values()


@mark.parametrize(
    "nodelist,scontrol_output,actual_lead_host",
    [
        ("host[1,2,3,4]", "host1\nhost2\nhost3\nhost4\n", "host1"),
        ("host1", "host1", "host1"),
        (
            "host-1-[1-3,5],host-2-[1,4]",
            "host-1-1\nhost-1-2\nhost-1-3\nhost-1-5\nhost-2-1\nhost-2-4\n",
            "host-1-1",
        ),
    ],
)
@mark.asyncio
@mock.patch("lm_agent.workload_managers.slurm.cmd_utils.run_command")
async def test_get_lead_host__success(
    run_command_mock: mock.MagicMock, nodelist: str, scontrol_output: str, actual_lead_host: str
):
    """
    Do I return the correct lead host from the scontrol show hostnames command?
    """
    run_command_mock.return_value = scontrol_output

    parsed_lead_host = await get_lead_host(nodelist)
    assert parsed_lead_host == actual_lead_host


@mark.asyncio
@mock.patch("lm_agent.workload_managers.slurm.cmd_utils.run_command")
async def test_get_lead_host__raise_exception_when_empty(run_command_mock: mock.MagicMock):
    """
    Do I raise an exception when the scontrol show hostnames command returns an empty string?
    """
    run_command_mock.return_value = ""

    with raises(ScontrolRetrievalFailure):
        await get_lead_host("host1")



"""
Test the OLicense parser.
"""

from lm_agent.models import LicenseUsesItem
from lm_agent.parsing.olicense import parse, parse_feature_line, parse_in_use_line, parse_usage_line


def test_parse_feature_line():
    """
    Does the regex for the feature line match the lines in the output?
    The line contains:
    - feature
    - license type
    - total
    - expiration date
    From these, we only need to extract ``feature`` and ``total``.
    """
    assert parse_feature_line("  ftire_adams;         	FreeFloating;	3;	2022-12-31 23:59:59;") == {
        "feature": "ftire_adams",
        "total": 3,
    }
    assert parse_feature_line("  ftire_adams;         	FreeFloating;	1;	2023-02-28 23:59:00;") == {
        "feature": "ftire_adams",
        "total": 1,
    }
    assert parse_feature_line("not a feature line") is None
    assert parse_feature_line("Licenser:	cosin scientific software") is None
    assert parse_feature_line("") is None


def test_parse_in_use_line():
    """
    Does the regex for the in use line match the lines in the output?
    The line contains:
    - in_use
    """
    assert parse_in_use_line("    2 FloatsLockedBy:") == 2
    assert parse_in_use_line("    1 FloatsLockedBy 'unknown clients'") == 1
    assert parse_in_use_line("not a in use license") is None
    assert parse_in_use_line("") is None


def test_parse_usage_line():
    """
    Does the regex for the usage line match the line in the output?
    The line contains:
    - user name
    - lead host
    - booked
    """
    assert parse_usage_line("        sbhyma@RD0087712 #1") == LicenseUsesItem(
        username="sbhyma",
        lead_host="RD0087712",
        booked=1,
    )
    assert parse_usage_line("        sbhyma@p-c39.maas.rnd.com #1") == LicenseUsesItem(
        username="sbhyma",
        lead_host="p-c39.maas.rnd.com",
        booked=1,
    )
    assert parse_usage_line("not a usage line") is None
    assert parse_usage_line("") is None


def test_parse__correct_output(olicense_output):
    """
    Does the parser return the correct data for this output?
    - olicense_output: expected output from the license server,
    which contain licenses and usage information.
    """
    assert parse(olicense_output) == {
        "ftire_adams": {
            "total": 4,
            "used": 3,
            "uses": [
                LicenseUsesItem(username="sbhyma", lead_host="RD0087712", booked=1),
                LicenseUsesItem(username="sbhyma", lead_host="RD0087713", booked=1),
                LicenseUsesItem(username="user22", lead_host="RD0087713", booked=1),
            ],
        }
    }


def test_parse__bad_output(olicense_output_bad):
    """
    Does the parser return the correct data for this output?
    - olicense_output_bad: unparseable output from the license server,
    which can happen when a connection error occours.
    """
    assert parse(olicense_output_bad) == {}


def test_parse__no_licenses_output(olicense_output_no_licenses):
    """
    Does the parser return the correct data for this output?
    - olicense_output_no_licenses: expected output from the server
    when none of the licenses are in use by users.
    """
    assert parse(olicense_output_no_licenses) == {"ftire_adams": {"total": 4, "used": 0, "uses": []}}



"""
Test the LM-X license server interface.
"""

from unittest import mock

from pytest import fixture, mark, raises

from lm_agent.config import settings
from lm_agent.exceptions import CommandFailedToExecute, LicenseManagerBadServerOutput
from lm_agent.models import LicenseReportItem
from lm_agent.server_interfaces.lmx import LMXLicenseServer


@fixture
def lmx_server(one_configuration_row_lmx) -> LMXLicenseServer:
    return LMXLicenseServer(one_configuration_row_lmx.license_servers)


def test_get_lmx_commands_list(lmx_server: LMXLicenseServer):
    """
    Do the commands for invoking the license server have the correct data?
    """
    commands_list = lmx_server.get_commands_list()
    assert commands_list == [
        [
            f"{settings.LMXENDUTIL_PATH}",
            "-licstat",
            "-host",
            "127.0.0.1",
            "-port",
            "2345",
        ]
    ]


@mark.asyncio
@mock.patch("lm_agent.server_interfaces.lmx.run_command")
async def test_lmx_get_output_from_server(
    run_command_mock: mock.MagicMock,
    lmx_server: LMXLicenseServer,
    lmx_output: str,
):
    """
    Do the license server interface return the output from the license server?
    """
    run_command_mock.return_value = lmx_output

    output = await lmx_server.get_output_from_server()
    assert output == lmx_output


@mark.asyncio
@mock.patch("lm_agent.server_interfaces.lmx.LMXLicenseServer.get_output_from_server")
async def test_lmx_get_report_item(
    get_output_from_server_mock: mock.MagicMock, lmx_server: LMXLicenseServer, lmx_output: str
):
    """
    Do the LM-X server interface generate a report item for the product?
    """
    get_output_from_server_mock.return_value = lmx_output

    assert await lmx_server.get_report_item(1, "hyperworks.hyperworks") == LicenseReportItem(
        feature_id=1,
        product_feature="hyperworks.hyperworks",
        used=25000,
        total=1000000,
        uses=[
            {"username": "sssaah", "lead_host": "RD0082406", "booked": 25000},
        ],
    )


@mark.asyncio
@mock.patch("lm_agent.server_interfaces.lmx.LMXLicenseServer.get_output_from_server")
async def test_lmx_get_report_item_with_bad_output(
    get_output_from_server_mock: mock.MagicMock, lmx_server: LMXLicenseServer, lmx_output_bad: str
):
    """
    Do the LM-X server interface raise an exception when the server returns an unparseable output?
    """
    get_output_from_server_mock.return_value = lmx_output_bad

    with raises(LicenseManagerBadServerOutput):
        await lmx_server.get_report_item(1, "hyperworks.hyperworks")


@mark.asyncio
@mock.patch("lm_agent.server_interfaces.lmx.LMXLicenseServer.get_output_from_server")
async def test_lmx_get_report_item_with_no_used_licenses(
    get_output_from_server_mock: mock.MagicMock,
    lmx_server: LMXLicenseServer,
    lmx_output_no_licenses: str,
):
    """
    Do the LM-X server interface generate a report item when no licenses are in use?
    """
    get_output_from_server_mock.return_value = lmx_output_no_licenses

    assert await lmx_server.get_report_item(1, "hyperworks.hyperworks") == LicenseReportItem(
        feature_id=1,
        product_feature="hyperworks.hyperworks",
        used=0,
        total=1000000,
        used_licenses=[],
    )


@mark.asyncio
@mock.patch("lm_agent.server_interfaces.lmx.run_command")
@mock.patch("lm_agent.server_interfaces.lmx.LMXLicenseServer.get_commands_list")
async def test_lmx_get_report_item_continues_on_exception(
    get_commands_list_mock: mock.MagicMock,
    run_command_mock: mock.MagicMock,
    lmx_server: LMXLicenseServer,
    lmx_output: str,
):
    """
    Do the LM-X server interface check the next command if the previous one fails?
    """
    get_commands_list_mock.return_value = [
        [
            f"{settings.LMXENDUTIL_PATH}",
            "-licstat",
            "-host",
            "127.0.0.1",
            "-port",
            "2345",
        ],
        [
            f"{settings.LMXENDUTIL_PATH}",
            "-licstat",
            "-host",
            "127.0.0.1",
            "-port",
            "3456",
        ],
    ]

    run_command_mock.side_effect = [CommandFailedToExecute("Command failed for first server"), lmx_output]

    assert await lmx_server.get_report_item(1, "hyperworks.hyperworks") == LicenseReportItem(
        feature_id=1,
        product_feature="hyperworks.hyperworks",
        used=25000,
        total=1000000,
        uses=[
            {"username": "sssaah", "lead_host": "RD0082406", "booked": 25000},
        ],
    )



"""
Test Slurm reservations.
"""

from unittest import mock

from pytest import fixture, mark

from lm_agent.config import settings
from lm_agent.exceptions import CommandFailedToExecute
from lm_agent.workload_managers.slurm.reservations import (
    create_or_update_reservation,
    scontrol_create_reservation,
    scontrol_delete_reservation,
    scontrol_show_reservation,
    scontrol_update_reservation,
)


@fixture
def reservation_license_data():
    return "test.license@licenseserver:10,another.license@server:25"


@fixture
def reservation_duration():
    return "24:00:00"


@fixture
def reservation_show_output():
    return """
    ReservationName=license-manager-reservation StartTime=2023-01-11T02:38:16 EndTime=2023-01-12T02:38:16 Duration=24:00:00
    Nodes=(null) NodeCnt=0 CoreCnt=0 Features=(null) PartitionName=(null) Flags=ANY_NODES
    TRES=(null)
    Users=license-manager Groups=(null) Accounts=(null) Licenses=test.license@licenseserver:10,another.license@server:25 State=ACTIVE BurstBuffer=(null) Watts=n/a
    MaxStartDelay=(null)
    """


@mark.asyncio
@mock.patch("lm_agent.workload_managers.slurm.reservations.run_command")
async def test__scontrol_create_reservation__success(
    run_command_mock: mock.MagicMock, reservation_license_data, reservation_duration
):
    """Test if a reservation is created succesfully."""
    run_command_mock.return_value = f"Reservation created: {settings.RESERVATION_IDENTIFIER}"
    create_reservation = await scontrol_create_reservation(reservation_license_data, reservation_duration)
    assert create_reservation


@mark.asyncio
@mock.patch("lm_agent.workload_managers.slurm.reservations.run_command")
async def test__scontrol_create_reservation__fail(
    run_command_mock: mock.MagicMock, reservation_license_data, reservation_duration
):
    """Test if reservation creation fails when the command fails or the reservation is invalid."""
    run_command_mock.side_effect = CommandFailedToExecute("Command failed")
    create_reservation = await scontrol_create_reservation(reservation_license_data, reservation_duration)
    assert not create_reservation


@mark.asyncio
@mock.patch("lm_agent.workload_managers.slurm.reservations.run_command")
async def test__scontrol_show_reservation__success(run_command_mock: mock.MagicMock, reservation_show_output):
    """Test if reservation is fetched succesfully."""
    run_command_mock.return_value = reservation_show_output
    read_reservation = await scontrol_show_reservation()
    assert read_reservation == reservation_show_output


@mark.asyncio
@mock.patch("lm_agent.workload_managers.slurm.reservations.run_command")
async def test__scontrol_show_reservation__fail(run_command_mock: mock.MagicMock):
    """Test if reservation fetch fails when the command fails or the reservation is not found."""
    run_command_mock.side_effect = CommandFailedToExecute("Command failed")
    read_reservation = await scontrol_show_reservation()
    assert not read_reservation


@mark.asyncio
@mock.patch("lm_agent.workload_managers.slurm.reservations.run_command")
async def test__scontrol_update_reservation__success(
    run_command_mock: mock.MagicMock, reservation_license_data, reservation_duration
):
    """Test if reservation is updated succesfully."""
    run_command_mock.return_value = "Reservation updated."
    update_reservation = await scontrol_update_reservation(reservation_license_data, reservation_duration)
    assert update_reservation


@mark.asyncio
@mock.patch("lm_agent.workload_managers.slurm.reservations.run_command")
async def test__scontrol_update_reservation__fail(
    run_command_mock: mock.MagicMock, reservation_license_data, reservation_duration
):
    """Test if reservation update fails when the command fails or the reservation is invalid."""
    run_command_mock.side_effect = CommandFailedToExecute("Command failed")
    update_reservation = await scontrol_update_reservation(reservation_license_data, reservation_duration)
    assert not update_reservation


@mark.asyncio
@mock.patch("lm_agent.workload_managers.slurm.reservations.run_command")
async def test__scontrol_delete_reservation__success(run_command_mock: mock.MagicMock):
    """Test if reservation is deleted succesfully."""
    run_command_mock.return_value = ""
    delete_reservation = await scontrol_delete_reservation()
    assert delete_reservation


@mark.asyncio
@mock.patch("lm_agent.workload_managers.slurm.reservations.run_command")
async def test__scontrol_delete_reservation__fail(run_command_mock: mock.MagicMock):
    """Test if reservation deletion fails when the command fails or the reservation is invalid."""
    run_command_mock.side_effect = CommandFailedToExecute("Command failed")
    delete_reservation = await scontrol_delete_reservation()
    assert not delete_reservation


@mark.asyncio
@mock.patch("lm_agent.workload_managers.slurm.reservations.scontrol_create_reservation")
@mock.patch("lm_agent.workload_managers.slurm.reservations.scontrol_show_reservation")
@mock.patch("lm_agent.workload_managers.slurm.reservations.scontrol_update_reservation")
@mock.patch("lm_agent.workload_managers.slurm.reservations.scontrol_delete_reservation")
async def test_create_or_update_reservation(delete_mock, update_mock, show_mock, create_mock):
    """
    Test that create_or_update_reservation:
    - update the reservation if it exists
    - delete the reservation if it can't update
    - create the reservation if doesn't exist
    """
    # Update reservation if it exists
    show_mock.return_value = "reservation_data"
    await create_or_update_reservation("reservation_info")
    update_mock.assert_called_once()

    # Delete reservation if it can't update
    show_mock.return_value = "reservation_data"
    update_mock.return_value = False
    await create_or_update_reservation("reservation_info")
    delete_mock.assert_called()

    # Create reservation if it doesn't exist
    show_mock.return_value = False
    await create_or_update_reservation("reservation_info")
    create_mock.assert_called()



import stat
from datetime import datetime, timezone
from unittest import mock

import jwt
import pytest
from httpx import Response
from pytest import mark, raises

from lm_agent.backend_utils.utils import (
    TOKEN_FILE_NAME,
    _load_token_from_cache,
    _write_token_to_cache,
    acquire_token,
    check_backend_health,
    get_all_features_bookings_sum,
    get_cluster_configs_from_backend,
    get_cluster_jobs_from_backend,
    make_booking_request,
    make_feature_update,
    remove_job_by_slurm_job_id,
    report_cluster_status,
)
from lm_agent.config import settings
from lm_agent.exceptions import LicenseManagerBackendConnectionError
from lm_agent.models import (
    BookingSchema,
    ConfigurationSchema,
    FeatureSchema,
    JobSchema,
    LicenseBooking,
    LicenseBookingRequest,
    LicenseServerSchema,
    LicenseServerType,
    ProductSchema,
)


def test__write_token_to_cache__caches_a_token(mock_cache_dir):
    """
    Verifies that the auth token can be saved in the cache directory.
    """
    mock_cache_dir.mkdir()
    _write_token_to_cache("dummy-token")
    token_path = mock_cache_dir / TOKEN_FILE_NAME
    assert token_path.exists()
    assert token_path.read_text() == "dummy-token"

    # Assert that the permissions for the file are read/write only for the current user
    assert token_path.stat().st_mode & (stat.S_IRWXU | stat.S_IRWXG | stat.S_IRWXO) == 0o600


def test__write_token_to_cache__warns_if_cache_dir_does_not_exist(caplog, mock_cache_dir):
    """
    Verifies that a warning is logged if the cache dire does not exists and that a token file is not created.
    """
    _write_token_to_cache("dummy-token")
    assert "Couldn't create missing cache directory" not in caplog.text

    token_path = mock_cache_dir / TOKEN_FILE_NAME
    assert not token_path.exists()


def test__write_token_to_cache__warns_if_token_cannot_be_created(caplog, mock_cache_dir):
    """
    Verifies that a warning is logged if the token file cannot be written.
    """
    mock_cache_dir.mkdir()
    token_path = mock_cache_dir / TOKEN_FILE_NAME
    token_path.write_text("pre-existing data")
    # Make file inaccessible permission-wise
    token_path.chmod(0o000)

    _write_token_to_cache("dummy-token")

    # Make file read/write again
    token_path.chmod(0o600)
    assert token_path.exists()
    assert token_path.read_text() == "pre-existing data"
    assert "Couldn't save token" in caplog.text


def test__load_token_from_cache__loads_token_data_from_the_cache(mock_cache_dir):
    """
    Verifies that a token can be retrieved from the cache.
    """
    mock_cache_dir.mkdir()
    token_path = mock_cache_dir / TOKEN_FILE_NAME
    one_minute_from_now = int(datetime.now(tz=timezone.utc).timestamp()) + 60
    created_token = jwt.encode(
        dict(exp=one_minute_from_now),
        key="dummy-key",
        algorithm="HS256",
    )
    token_path.write_text(created_token)
    retrieved_token = _load_token_from_cache()
    assert retrieved_token == created_token


def test__load_token_from_cache__returns_none_if_cached_token_does_not_exist():
    """
    Verifies that None is returned if the cached token does not exist.
    """
    retrieved_token = _load_token_from_cache()
    assert retrieved_token is None


def test__load_token_from_cache__returns_none_and_warns_if_cached_token_cannot_be_read(
    caplog, mock_cache_dir
):
    """
    Verifies that None is returned if the token cannot be read. Also checks that a warning is logged.
    """
    mock_cache_dir.mkdir()
    token_path = mock_cache_dir / TOKEN_FILE_NAME
    one_minute_from_now = int(datetime.now(tz=timezone.utc).timestamp()) + 60
    created_token = jwt.encode(
        dict(exp=one_minute_from_now),
        key="dummy-key",
        algorithm="HS256",
    )
    token_path.write_text(created_token)
    token_path.chmod(0o000)

    retrieved_token = _load_token_from_cache()
    assert retrieved_token is None
    assert "Couldn't load token" in caplog.text


def test__load_token_from_cache__returns_none_and_warns_if_cached_token_is_expired(caplog, mock_cache_dir):
    """
    Verifies that None is returned if the token is expired. Also checks that a warning is logged.
    """
    mock_cache_dir.mkdir()
    token_path = mock_cache_dir / TOKEN_FILE_NAME
    one_second_ago = int(datetime.now(tz=timezone.utc).timestamp()) - 1
    expired_token = jwt.encode(dict(exp=one_second_ago), key="dummy-key", algorithm="HS256")
    token_path.write_text(expired_token)

    retrieved_token = _load_token_from_cache()
    assert retrieved_token is None
    assert "Cached token is expired" in caplog.text


def test__load_token_from_cache__returns_none_and_warns_if_cached_token_will_expire_soon(
    caplog, mock_cache_dir
):
    """
    Verifies that None is returned if the token will expired soon. Also checks that a warning is logged.
    """
    mock_cache_dir.mkdir()
    token_path = mock_cache_dir / TOKEN_FILE_NAME
    nine_seconds_from_now = int(datetime.now(tz=timezone.utc).timestamp()) + 9
    expired_token = jwt.encode(dict(exp=nine_seconds_from_now), key="dummy-key", algorithm="HS256")
    token_path.write_text(expired_token)

    retrieved_token = _load_token_from_cache()
    assert retrieved_token is None
    assert "Cached token is expired" in caplog.text


def test_acquire_token__gets_a_token_from_the_cache(mock_cache_dir):
    """
    Verifies that the token is retrieved from the cache if it is found there.
    """
    mock_cache_dir.mkdir()
    token_path = mock_cache_dir / TOKEN_FILE_NAME
    one_minute_from_now = int(datetime.now(tz=timezone.utc).timestamp()) + 60
    created_token = jwt.encode(
        dict(exp=one_minute_from_now),
        key="dummy-key",
        algorithm="HS256",
    )
    token_path.write_text(created_token)
    retrieved_token = acquire_token()
    assert retrieved_token == created_token


def test_acquire_token__gets_a_token_from_auth_0_if_one_is_not_in_the_cache(respx_mock, mock_cache_dir):
    """
    Verifies that a token is pulled from OIDC if it is not found in the cache.
    Also checks to make sure the token is cached.
    """
    mock_cache_dir.mkdir()
    token_path = mock_cache_dir / TOKEN_FILE_NAME
    assert not token_path.exists()

    retrieved_token = acquire_token()
    assert retrieved_token == "dummy-token"

    assert token_path.read_text() == retrieved_token


@mark.asyncio
@pytest.mark.respx(base_url=str(settings.BACKEND_BASE_URL))
async def test__check_backend_health__success_on_two_hundered(respx_mock):
    respx_mock.get("/lm/health").mock(return_value=Response(204))
    await check_backend_health()


@mark.asyncio
@pytest.mark.respx(base_url=str(settings.BACKEND_BASE_URL))
async def test__get_license_manager_backend_version__raises_exception_on_non_two_hundred(respx_mock):
    respx_mock.get("/lm/health").mock(return_value=Response(500))
    with raises(LicenseManagerBackendConnectionError, match="Could not connect"):
        await check_backend_health()


@mark.asyncio
@pytest.mark.respx(base_url="http://backend")
async def test__report_cluster_status__success_on_202(respx_mock):
    respx_mock.put("/lm/cluster_statuses?interval=60").mock(return_value=Response(202))
    await report_cluster_status()


@mark.asyncio
@pytest.mark.respx(base_url="http://backend")
async def test__report_cluster_status__raises_exception_on_non_202(respx_mock):
    respx_mock.put("/lm/cluster_statuses?interval=60").mock(return_value=Response(400))

    with raises(LicenseManagerBackendConnectionError, match="Failed to report cluster status"):
        await report_cluster_status()


@mark.asyncio
@pytest.mark.respx(base_url="http://backend")
async def test__get_cluster_jobs_from_backend(jobs, respx_mock):
    """Test that get_jobs_from_backend parses and returns the jobs from the cluster."""
    respx_mock.get("/lm/jobs/by_client_id").mock(
        return_value=Response(
            status_code=200,
            json=jobs[:2],
        )
    )

    expected_jobs = [
        JobSchema(
            id=1,
            slurm_job_id="123",
            cluster_client_id="dummy",
            username="user1",
            lead_host="host1",
            bookings=[
                BookingSchema(id=1, job_id=1, feature_id=1, quantity=12),
                BookingSchema(id=2, job_id=1, feature_id=2, quantity=50),
            ],
        ),
        JobSchema(
            id=2,
            slurm_job_id="456",
            cluster_client_id="dummy",
            username="user2",
            lead_host="host2",
            bookings=[
                BookingSchema(id=3, job_id=2, feature_id=4, quantity=15),
                BookingSchema(id=4, job_id=2, feature_id=7, quantity=25),
            ],
        ),
    ]

    jobs = await get_cluster_jobs_from_backend()
    assert jobs == expected_jobs


@mark.asyncio
@pytest.mark.respx(base_url="http://backend")
async def test__get_cluster_configs_from_backend(configurations, respx_mock):
    """Test that get_configs_from_backend parses and returns the configurations from the cluster."""
    respx_mock.get("/lm/configurations/by_client_id").mock(
        return_value=Response(
            status_code=200,
            json=configurations,
        )
    )

    expected_configs = [
        ConfigurationSchema(
            id=1,
            name="Abaqus",
            cluster_client_id="dummy",
            features=[
                FeatureSchema(
                    id=1,
                    name="abaqus",
                    product=ProductSchema(id=1, name="abaqus"),
                    config_id=1,
                    reserved=100,
                    total=123,
                    used=12,
                    booked_total=12,
                )
            ],
            license_servers=[
                LicenseServerSchema(id=1, config_id=1, host="licserv0001", port=1234),
                LicenseServerSchema(id=3, config_id=1, host="licserv0003", port=8760),
            ],
            grace_time=60,
            type=LicenseServerType.FLEXLM,
        ),
        ConfigurationSchema(
            id=2,
            name="Converge",
            cluster_client_id="dummy",
            features=[
                FeatureSchema(
                    id=2,
                    name="converge_super",
                    product=ProductSchema(id=2, name="converge"),
                    config_id=2,
                    reserved=0,
                    total=500,
                    used=50,
                    booked_total=50,
                )
            ],
            license_servers=[
                LicenseServerSchema(id=2, config_id=2, host="licserv0002", port=2345),
            ],
            grace_time=123,
            type=LicenseServerType.RLM,
        ),
    ]

    configs = await get_cluster_configs_from_backend()
    assert configs == expected_configs


@pytest.mark.asyncio
@mock.patch("lm_agent.backend_utils.utils.get_all_features_from_backend")
async def test__get_all_features_bookings_sum(get_all_features_mock, parsed_features):
    """
    Test that get_bookings_sum_per_cluster returns the booking sum per cluster for a given product_feature.
    """
    get_all_features_mock.return_value = parsed_features
    bookings_sum = await get_all_features_bookings_sum()
    assert bookings_sum["abaqus.abaqus"] == 62


@pytest.mark.asyncio
@pytest.mark.respx(base_url="http://backend")
async def test__make_feature_update__success(respx_mock):
    """
    Test that make_feature_update updates the features correctly.
    """
    features_to_update = [
        {
            "product_name": "abaqus",
            "feature_name": "abaqus",
            "total": 500,
            "used": 50,
        },
        {
            "product_name": "converge",
            "feature_name": "converge_super",
            "total": 100,
            "used": 10,
        },
    ]

    route = respx_mock.put("/lm/features/bulk").mock(return_value=Response(status_code=200))

    await make_feature_update(features_to_update)
    assert route.called


@pytest.mark.asyncio
@pytest.mark.respx(base_url="http://backend")
async def test__make_feature_update__raises_exception_on_non_two_hundred(respx_mock):
    """
    Test that make_feature_update handles a failed feature update correctly.
    """
    features_to_update = [
        {
            "product_name": "abaqus",
            "feature_name": "abaqus",
            "total": 500,
            "used": 50,
        },
        {
            "product_name": "converge",
            "feature_name": "converge_super",
            "total": 100,
            "used": 10,
        },
    ]

    respx_mock.put("/lm/features/bulk").mock(return_value=Response(status_code=500))

    with pytest.raises(LicenseManagerBackendConnectionError):
        await make_feature_update(features_to_update)


@pytest.mark.asyncio
@pytest.mark.respx(base_url="http://backend")
async def test__make_booking_request__success(respx_mock):
    """
    Test that make_booking_request successfully creates a job and its bookings on the backend.
    """
    lbr = LicenseBookingRequest(
        slurm_job_id="12345",
        username="test_user",
        lead_host="test_host",
        bookings=[
            LicenseBooking(product_feature="abaqus.abaqus", quantity=5),
            LicenseBooking(product_feature="converge.converge_super", quantity=10),
        ],
    )

    respx_mock.post("/lm/jobs").mock(
        return_value=Response(
            status_code=201,
            json={"id": 1},
        )
    )

    result = await make_booking_request(lbr)
    assert result is True


@pytest.mark.asyncio
@pytest.mark.respx(base_url="http://backend")
async def test__make_booking_request_job__returns_false_on_booking_failure(respx_mock):
    """
    Test that make_booking_request handles the failure case when booking creation fails.
    """
    lbr = LicenseBookingRequest(
        slurm_job_id="12345",
        username="test_user",
        lead_host="test_host",
        bookings=[
            LicenseBooking(product_feature="abaqus.abaqus", quantity=5),
        ],
    )

    respx_mock.post("/lm/jobs").mock(
        return_value=Response(
            status_code=409,
            json={"message": "Not enough licenses"},
        )
    )

    assert not await make_booking_request(lbr)


@pytest.mark.asyncio
@pytest.mark.respx(base_url="http://backend")
async def test__remove_job_by_slurm_job_id__success(respx_mock):
    """
    Test that remove_job_by_slurm_job_id successfully removes the job in the cluster.
    """
    slurm_job_id = "12345"

    route = respx_mock.delete(f"/lm/jobs/slurm_job_id/{slurm_job_id}").mock(
        return_value=Response(
            status_code=200,
            json={"message": "Job removed successfully"},
        )
    )

    await remove_job_by_slurm_job_id(slurm_job_id)
    assert route.called


@pytest.mark.asyncio
@pytest.mark.respx(base_url="http://backend")
async def test__remove_job_by_slurm_job_id__raises_exception_on_non_two_hundred(respx_mock):
    """
    Test that remove_job_by_slurm_job_id raises an exception when the job removal fails.
    """
    slurm_job_id = "12345"

    respx_mock.delete(f"/lm/jobs/slurm_job_id/{slurm_job_id}").mock(
        return_value=Response(
            status_code=500,
            json={"error": "Internal Server Error"},
        )
    )

    with pytest.raises(LicenseManagerBackendConnectionError):
        await remove_job_by_slurm_job_id(slurm_job_id)
